#ifndef DUKE_H
#define DUKE_H

#include "gbTypes.h"

class dkMaterial {
public:
								dkMaterial() {}
	virtual						~dkMaterial() {}

	virtual void				DrawSliver(BITMAP* pDest, gbInt dwX, gbInt dwY1, gbInt dwY2, gbFloat fU, gbFloat fV1, gbFloat fV2) = 0;

protected:
	gbInt						m_dwColor;

};
typedef dkMaterial *dkMaterial_p;

class dkFlatMaterial : public dkMaterial {
public:
	dkFlatMaterial() : m_dwColor(0) {}
	virtual ~dkFlatMaterial() {}

	virtual void DrawSliver(BITMAP* pDest, gbInt dwX, gbInt dwY1, gbInt dwY2, gbFloat fU, gbFloat fV1, gbFloat fV2)
	{
		if(dwY2 >= dwY1)
			vline(pDest, dwX, dwY1, dwY2, m_dwColor);
	}

public: // Inlines
	inline gbInt GetColor()
	{
		return m_dwColor;
	}

	inline void SetColor(gbInt dwValue)
	{
		m_dwColor = dwValue;
	}

protected:
	gbInt						m_dwColor;

};
typedef dkFlatMaterial *dkFlatMaterial_p;

class dkTexMaterial : public dkMaterial {
public:
								dkTexMaterial();
	virtual						~dkTexMaterial();

	virtual void				DrawSliver(BITMAP* pDest, gbInt dwX, gbInt dwY1, gbInt dwY2, gbFloat fU, gbFloat fV1, gbFloat fV2);

	virtual bool				Load(const gbString& sFile1,
									const gbString& sFile2,
									const gbString& sFile3,
									const gbString& sFile4,
									const gbString& sFile5,
									const gbString& sFile6);

protected:
	gbList(gbInt*)				m_pvSlivers1;
	gbList(gbInt*)				m_pvSlivers2;
	gbList(gbInt*)				m_pvSlivers3;
	gbList(gbInt*)				m_pvSlivers4;
	gbList(gbInt*)				m_pvSlivers5;
	gbList(gbInt*)				m_pvSlivers6;
	gbInt						m_dwvSliverSize[6];

};
typedef dkTexMaterial *dkTexMaterial_p;

class dkVertex {
public:
								dkVertex() :
									m_fX(0.0f),
									m_fY(0.0f)
								{}
								~dkVertex() {}

public: // Inlines
	inline gbFloat GetX()
	{
		return m_fX;
	}
	inline gbFloat GetY()
	{
		return m_fY;
	}

	inline void SetX(gbFloat fValue)
	{
		m_fX = fValue;
	}
	inline void SetY(gbFloat fValue)
	{
		m_fY = fValue;
	}

protected:
	gbFloat						m_fX;
	gbFloat						m_fY;

};
typedef dkVertex *dkVertex_p;

class dkWall {
public:
								dkWall();
								~dkWall();

	void						SetVertices(dkVertex_p pStart, dkVertex_p pEnd);
	void						Join(dkWall* pBackWall);

	gbFloat						PointPosition(gbFloat fX, gbFloat fY);
	gbFloat						PointOffset(gbFloat fX, gbFloat fY);
	void						PointClosest(gbFloat fX, gbFloat fY, gbFloat& fOutX, gbFloat& fOutY);
	gbFloat						RayIntersect(gbFloat fX, gbFloat fY, gbFloat fDX, gbFloat fDY, gbFloat& fOutX, gbFloat& fOutY);

	friend class				dkMap;
	friend class				dkSector;

public: // Inlines
	inline dkVertex_p GetStart()
	{
		return m_pStart;
	}
	inline dkVertex_p GetEnd()
	{
		return m_pEnd;
	}
	inline class dkWall* GetBackWall()
	{
		return m_pBackWall;
	}
	inline class dkWall* GetPrev()
	{
		return m_pPrev;
	}
	inline class dkWall* GetNext()
	{
		return m_pNext;
	}
	inline dkMaterial_p GetMaterial()
	{
		return m_pMaterial;
	}

	inline void SetBackWall(class dkWall* pSector)
	{
		m_pBackWall = pSector;
	}
	inline void SetMaterial(dkMaterial_p pMaterial)
	{
		m_pMaterial = pMaterial;
	}

	inline gbFloat GetLength()
	{
		return m_fLength;
	}
	inline const gbFloat* GetDirection()
	{
		return m_fvDirection;
	}
	inline const gbFloat* GetNormal()
	{
		return m_fvNormal;
	}

protected:
	dkVertex_p					m_pStart;
	dkVertex_p					m_pEnd;
	class dkWall*				m_pBackWall;
	dkMaterial_p				m_pMaterial;

	gbFloat						m_fvDirection[2];
	gbFloat						m_fvNormal[2];
	gbFloat						m_fLength;

	dkWall*						m_pPrev;
	dkWall*						m_pNext;
	class dkMap*				m_pMap;
	class dkSector*				m_pSector;

};
typedef dkWall *dkWall_p;

class dkSector {
public:
								dkSector();
								~dkSector();

	dkWall_p					AddWallPair(dkVertex_p pStart, dkVertex_p pEnd);
	dkWall_p					SplitWall(dkWall_p pWall, dkVertex_p pVertex);
	void						DeleteWall(dkWall_p pWall);
	void						UpdateWalls();

	gbBool						IsPointInside(gbFloat fX, gbFloat fY);
	gbFloat						RayEntryPoint(gbFloat fX, gbFloat fY, gbFloat fDX, gbFloat fDY, gbFloat& fOutX, gbFloat& fOutY);

	void						Draw(BITMAP* pDest, gbInt dwColumn, gbInt dwCH, gbInt dwFH, const gbFloat* fvRay, int dwBounceCount);

	friend class				dkMap;

public: // Inlines
	inline gbInt CountWalls()
	{
		return m_pvWalls.size();
	}
	inline dkWall_p GetWall(gbInt dwIndex)
	{
		return m_pvWalls[dwIndex];
	}

	inline gbFloat GetFloorHeight()
	{
		return m_fFloorHeight;
	}
	inline gbFloat GetCeilingHeight()
	{
		return m_fCeilingHeight;
	}
	inline gbFloat GetSlopeOffset(gbFloat fX, gbFloat fY)
	{
		return (fX-m_pvWalls[0]->m_pStart->GetX())*m_pvWalls[0]->m_fvNormal[0]+
				(fY-m_pvWalls[0]->m_pStart->GetY())*m_pvWalls[0]->m_fvNormal[1];
	}
	inline gbFloat GetFloorHeight(gbFloat fX, gbFloat fY)
	{
		return m_fFloorHeight+GetSlopeOffset(fX, fY)*m_fFloorSlope;
	}
	inline gbFloat GetCeilingHeight(gbFloat fX, gbFloat fY)
	{
		return m_fCeilingHeight+GetSlopeOffset(fX, fY)*m_fCeilingSlope;
	}
	inline gbFloat GetFloorSlope()
	{
		return m_fFloorSlope;
	}
	inline gbFloat GetCeilingSlope()
	{
		return m_fCeilingSlope;
	}
	inline dkMaterial_p GetFloorMaterial()
	{
		return m_pFloorMaterial;
	}
	inline dkMaterial_p GetCeilingMaterial()
	{
		return m_pCeilingMaterial;
	}

	inline void SetFloorHeight(gbFloat fValue)
	{
		m_fFloorHeight = fValue;
	}
	inline void SetCeilingHeight(gbFloat fValue)
	{
		m_fCeilingHeight = fValue;
	}
	inline void SetFloorSlope(gbFloat fValue)
	{
		m_fFloorSlope = fValue;
	}
	inline void SetCeilingSlope(gbFloat fValue)
	{
		m_fCeilingSlope = fValue;
	}
	inline void SetFloorMaterial(dkFlatMaterial_p pMaterial)
	{
		m_pFloorMaterial = pMaterial;
	}
	inline void SetCeilingMaterial(dkFlatMaterial_p pMaterial)
	{
		m_pCeilingMaterial = pMaterial;
	}

protected:
	gbList(dkWall_p)			m_pvWalls;
	gbFloat						m_fFloorHeight;
	gbFloat						m_fCeilingHeight;
	gbFloat						m_fFloorSlope;
	gbFloat						m_fCeilingSlope;
	dkFlatMaterial_p			m_pFloorMaterial;
	dkFlatMaterial_p			m_pCeilingMaterial;

	class dkMap*				m_pMap;

};
typedef dkSector *dkSector_p;

class dkMap {
public:
								dkMap();
								~dkMap();

	dkVertex_p					CreateVertex(gbFloat fX, gbFloat fY);
	dkSector_p					CreateSector();
	dkMaterial_p				CreateMaterial(gbInt dwR, gbInt dwG, gbInt dwB);
	dkMaterial_p				CreateMaterial(const gbString& sFile1,
									const gbString& sFile2,
									const gbString& sFile3,
									const gbString& sFile4,
									const gbString& sFile5,
									const gbString& sFile6);

	void						Draw(BITMAP* pDest, const gbFloat* fvEye);

public: // Inlines
	inline dkVertex_p GetVertex(gbInt dwIndex)
	{
		return m_pvVertices[dwIndex];
	}
	inline dkSector_p GetSector(gbInt dwIndex)
	{
		return m_pvSectors[dwIndex];
	}
	inline dkMaterial_p GetMaterial(gbInt dwIndex)
	{
		return m_pvMaterials[dwIndex];
	}

protected:
	gbList(dkVertex_p)			m_pvVertices;
	gbList(dkSector_p)			m_pvSectors;
	gbList(dkMaterial_p)		m_pvMaterials;

};
typedef dkMap *dkMap_p;

#endif // DUKE_H
