#include <cmath>
#define ALLEGRO_FIX_H  // Don't use Allegro's fix class
#include <allegro.h>
#include <winalleg.h>
#include "Duke.h"

void gbNormalise(gbFloat& fX, gbFloat& fY)
{
	gbFloat fLength = fX*fX+fY*fY;
	fX /= fLength;
	fY /= fLength;
}

dkMap::dkMap()
{
}

dkMap::~dkMap()
{
}

dkVertex_p dkMap::CreateVertex(gbFloat fX, gbFloat fY)
{
	dkVertex_p pVertex = new dkVertex();
	pVertex->SetX(fX);
	pVertex->SetY(fY);

	m_pvVertices.push_back(pVertex);

	return pVertex;
}

dkSector_p dkMap::CreateSector()
{
	dkSector_p pSector = new dkSector();
	pSector->m_pMap = this;

	m_pvSectors.push_back(pSector);

	return pSector;
}

dkMaterial_p dkMap::CreateMaterial(gbInt dwR, gbInt dwG, gbInt dwB)
{
	dkFlatMaterial_p pMaterial = new dkFlatMaterial();
	pMaterial->SetColor(makecol(dwR, dwG, dwB));

	m_pvMaterials.push_back(pMaterial);

	return pMaterial;
}

dkMaterial_p dkMap::CreateMaterial(const gbString& sFile1,
									const gbString& sFile2,
									const gbString& sFile3,
									const gbString& sFile4,
									const gbString& sFile5,
									const gbString& sFile6)
{
	dkTexMaterial_p pMaterial = new dkTexMaterial();
	pMaterial->Load(sFile1, sFile2, sFile3, sFile4, sFile5, sFile6);

	m_pvMaterials.push_back(pMaterial);

	return pMaterial;
}

void dkMap::Draw(BITMAP* pDest, const gbFloat* fvEye)
{
	dkSector_p pSector = NULL;
	gbFloat fvRay[4];
	gbFloat fPos;
	for(gbListIter(dkSector_p) iSec=m_pvSectors.begin(); iSec!=m_pvSectors.end(); ++iSec)
	{
		if((*iSec)  &&  (*iSec)->IsPointInside(fvEye[0], fvEye[1]))
		{
			pSector = (*iSec);
			break;
		}
	}

	if(pSector)
	{
		for(gbInt lp=0; lp<pDest->w; lp++)
		{
			fPos = lp/(gbFloat)(pDest->w-1);
			fvRay[0] = fvEye[0];
			fvRay[1] = fvEye[1];
			fvRay[2] = fvEye[2]+(fvEye[4]-fvEye[2])*fPos;
			fvRay[3] = fvEye[3]+(fvEye[5]-fvEye[3])*fPos;
			pSector->Draw(pDest, lp, 0, pDest->h-1, fvRay, 0);
		}
	}
	else
	{
		for(gbInt lp=0; lp<pDest->w; lp++)
		{
			fPos = lp/(gbFloat)(pDest->w-1);
			fvRay[0] = fvEye[0];
			fvRay[1] = fvEye[1];
			fvRay[2] = fvEye[2]+(fvEye[4]-fvEye[2])*fPos;
			fvRay[3] = fvEye[3]+(fvEye[5]-fvEye[3])*fPos;

			gbFloat fSX = fvRay[0];
			gbFloat fSY = fvRay[1];
			gbFloat fMinDist = 100000.0f;
			for(gbListIter(dkSector_p) iSec=m_pvSectors.begin(); iSec!=m_pvSectors.end(); ++iSec)
			{
				if((*iSec))
				{
					gbFloat fIX, fIY;
					gbFloat fDist = (*iSec)->RayEntryPoint(fvRay[0], fvRay[1], fvRay[2], fvRay[3], fIX, fIY);

					if(fDist > 0.0f  &&  fDist < fMinDist)
					{
						pSector = (*iSec);
						fSX = fIX;
						fSY = fIY;
						fMinDist = fDist;
					}
				}
			}

			if(pSector)
			{
				gbFloat fSlopePos = pSector->GetSlopeOffset(fSX, fSY);
				gbFloat fCeiling = pSector->m_fCeilingHeight+pSector->m_fCeilingSlope*fSlopePos;
				gbFloat fFloor = pSector->m_fFloorHeight+pSector->m_fFloorSlope*fSlopePos;
				gbInt dwTop = pDest->h/2-(int)(pDest->h*fCeiling/fMinDist);
				gbInt dwBottom = pDest->h/2-(int)(pDest->h*fFloor/fMinDist);
				pSector->Draw(pDest, lp, dwTop, dwBottom, fvRay, 0);
			}
		}
	}
}

/*
	dkWall_p pWall;
	gbFloat fX1, fY1, fX2, fY2;
		for(gbListIter(dkWall_p) iPos=(*iSec)->m_pvWalls.begin(); iPos!=(*iSec)->m_pvWalls.end(); ++iPos)
		{
			pWall = *iPos;
			fX1 = pWall->GetStart()->GetX();
			fY1 = pWall->GetStart()->GetY();
			fX2 = pWall->GetEnd()->GetX();
			fY2 = pWall->GetEnd()->GetY();
			line(pDest,
				(gbInt)(fX1*8)+SCREEN_W/2, (gbInt)(fY1*8)+SCREEN_H/2,
				(gbInt)(fX2*8)+SCREEN_W/2, (gbInt)(fY2*8)+SCREEN_H/2,
				pWall->m_pMaterial?pWall->m_pMaterial->GetColor():-1);

			fX1 = (fX1+fX2)/2.0f;
			fY1 = (fY1+fY2)/2.0f;
			fX2 = fX1+pWall->m_fvNormal[0]*2.0f;
			fY2 = fY1+pWall->m_fvNormal[1]*2.0f;
			line(pDest,
				(gbInt)(fX1*8)+SCREEN_W/2, (gbInt)(fY1*8)+SCREEN_H/2,
				(gbInt)(fX2*8)+SCREEN_W/2, (gbInt)(fY2*8)+SCREEN_H/2,
				pWall->m_pMaterial?pWall->m_pMaterial->GetColor():-1);
		}
*/
