#include <cmath>
#define ALLEGRO_FIX_H  // Don't use Allegro's fix class
#include <allegro.h>
#include <winalleg.h>
#include "Duke.h"

inline void dline(BITMAP* pDest, gbInt dwX, gbInt dwY1, gbInt dwY2, gbInt dwColor)
{
	if(dwY1 < dwY2)
		vline(pDest, dwX, dwY1, dwY2, dwColor);
}

inline void uline(BITMAP* pDest, gbInt dwX, gbInt dwY1, gbInt dwY2, gbInt dwColor)
{
	if(dwY1 > dwY2)
		vline(pDest, dwX, dwY1, dwY2, dwColor);
}

dkSector::dkSector() :
	m_fFloorHeight(-1.0f),
	m_fCeilingHeight(1.0f),
	m_fFloorSlope(0.0f),
	m_fCeilingSlope(0.0f),
	m_pFloorMaterial(NULL),
	m_pCeilingMaterial(NULL)
{
}

dkSector::~dkSector()
{
}

dkWall_p dkSector::AddWallPair(dkVertex_p pStart, dkVertex_p pEnd)
{
	dkWall_p pWall = new dkWall();
	pWall->m_pMap = m_pMap;
	pWall->m_pSector = this;
	pWall->SetVertices(pStart, pEnd);

	dkWall_p pWall2 = new dkWall();
	pWall2->m_pMap = m_pMap;
	pWall2->m_pSector = this;
	pWall2->SetVertices(pEnd, pStart);

	pWall->m_pPrev = pWall->m_pNext = pWall2;
	pWall2->m_pPrev = pWall2->m_pNext = pWall;

	m_pvWalls.push_back(pWall);
	m_pvWalls.push_back(pWall2);
	UpdateWalls();

	return pWall;
}

dkWall_p dkSector::SplitWall(dkWall_p pWall, dkVertex_p pVertex)
{
	dkWall_p pNewWall = new dkWall();
	pNewWall->m_pMap = m_pMap;
	pNewWall->m_pSector = this;
	pNewWall->SetVertices(pVertex, pWall->m_pEnd);

	pNewWall->m_pPrev = pWall;
	pNewWall->m_pNext = pWall->m_pNext;

	pWall->m_pNext->m_pPrev = pNewWall;
	pWall->m_pNext = pNewWall;
	pWall->SetVertices(pWall->m_pStart, pVertex);

	m_pvWalls.push_back(pNewWall);

	return pNewWall;
}

void dkSector::DeleteWall(dkWall_p pWall)
{
	if(!pWall)
		return;

	for(gbListIter(dkWall_p) iPos=m_pvWalls.begin(); iPos!=m_pvWalls.end(); )
	{
		if(*iPos)
		{
			iPos = m_pvWalls.erase(iPos);
		}
		else
		{
			++iPos;
		}
	}
	delete pWall;

	UpdateWalls();
}

void dkSector::UpdateWalls()
{
	for(gbListIter(dkWall_p) iPos=m_pvWalls.begin(); iPos!=m_pvWalls.end(); ++iPos)
	{
		(*iPos)->SetVertices((*iPos)->GetStart(), (*iPos)->GetEnd());
	}
}

gbBool dkSector::IsPointInside(gbFloat fX, gbFloat fY)
{
	bool bOdd = false;

    dkVertex_p pVert1, pVert2;
	for(gbListIter(dkWall_p) iPos=m_pvWalls.begin(); iPos!=m_pvWalls.end(); ++iPos)
	{
		pVert1 = (*iPos)->m_pStart;
		pVert2 = (*iPos)->m_pEnd;

		if(pVert1->GetY() < fY  &&  pVert2->GetY() >= fY  ||
			pVert2->GetY() < fY  &&  pVert1->GetY() >= fY)
		{
			if((pVert1->GetX()+(fY-pVert1->GetY())/(pVert2->GetY()-pVert1->GetY())*(pVert2->GetX()-pVert1->GetX())) < fX)
				bOdd = !bOdd;
		}
	}

	return bOdd;
}

gbFloat dkSector::RayEntryPoint(gbFloat fX, gbFloat fY, gbFloat fDX, gbFloat fDY, gbFloat& fOutX, gbFloat& fOutY)
{
	gbFloat fMinDist = 100000.0f;
	for(gbListIter(dkWall_p) iPos=m_pvWalls.begin(); iPos!=m_pvWalls.end(); ++iPos)
	{
		gbFloat fIX, fIY;
		gbFloat fIDist = (*iPos)->RayIntersect(fX, fY, fDX, fDY, fIX, fIY);
		gbFloat fIPos = (*iPos)->PointPosition(fIX, fIY);

		if(fIDist > 0.0f  &&  fIDist < fMinDist  &&  fIPos >= 0.0f  &&  fIPos < (*iPos)->m_fLength)
		{
			fMinDist = fIDist;
			fOutX = fIX+fDX*0.0125f;
			fOutY = fIY+fDY*0.0125f;
		}
	}

	return (fMinDist < 99999.0f)?fMinDist:-1.0f;
}

void dkSector::Draw(BITMAP* pDest, gbInt dwColumn, gbInt dwCH, gbInt dwFH, const gbFloat* fvRay, int dwBounceCount)
{
	if(dwCH > dwFH)
		return;

	dkWall_p pWall = NULL;
	gbFloat fWallPos = 0.0f;
	gbFloat fMinDist = 100000.0f;
	gbFloat fWallX, fWallY;
	for(gbListIter(dkWall_p) iPos=m_pvWalls.begin(); iPos!=m_pvWalls.end(); ++iPos)
	{
		if((*iPos)  &&  (fvRay[2]*(*iPos)->m_fvNormal[0]+fvRay[3]*(*iPos)->m_fvNormal[1]) < 0.0f)
		{
			gbFloat fIX, fIY;
			gbFloat fIDist = (*iPos)->RayIntersect(fvRay[0], fvRay[1], fvRay[2], fvRay[3], fIX, fIY);
			gbFloat fIPos = (*iPos)->PointPosition(fIX, fIY);

			if(fIDist > 0.0f  &&  fIDist < fMinDist  &&  fIPos >= 0.0f  &&  fIPos < (*iPos)->m_fLength)
			{
				fMinDist = fIDist;
				fWallX = fIX;
				fWallY = fIY;
				pWall = (*iPos);
				fWallPos = fIPos;
			}
		}
	}

	if(pWall)
	{
		gbFloat fSlopePos = GetSlopeOffset(fWallX, fWallY);
		if(pWall->m_pBackWall)
		{
			gbInt dwTop = pDest->h/2-(int)((pDest->h*(m_fCeilingHeight+m_fCeilingSlope*fSlopePos))/fMinDist);
			gbInt dwBottom = pDest->h/2-(int)(pDest->h*(m_fFloorHeight+m_fFloorSlope*fSlopePos)/fMinDist);
			dline(pDest, dwColumn, dwCH, dwTop, m_pCeilingMaterial?m_pCeilingMaterial->GetColor():0);
			uline(pDest, dwColumn, dwFH, dwBottom, m_pFloorMaterial?m_pFloorMaterial->GetColor():0);
			dkSector_p pBackSector = pWall->m_pBackWall->m_pSector;
			if(pBackSector)
			{
				gbFloat fBackSlopePos = pBackSector->GetSlopeOffset(fWallX, fWallY);
				gbInt dwBackTop = pDest->h/2-(int)(pDest->h*(pBackSector->m_fCeilingHeight+pBackSector->m_fCeilingSlope*fBackSlopePos)/fMinDist);
				gbInt dwBackBottom = pDest->h/2-(int)(pDest->h*(pBackSector->m_fFloorHeight+pBackSector->m_fFloorSlope*fBackSlopePos)/fMinDist);

				if(pWall->GetMaterial())
				{
					gbInt dwT = dwTop;
					gbInt dwB = dwBottom;
					dwT = MAX(dwT, dwCH);
					dwB = MIN(dwB, dwFH);
	
					gbFloat fV1 = (dwT-dwTop)/(gbFloat)(dwBottom-dwTop);
					gbFloat fV2 = (dwB-dwTop)/(gbFloat)(dwBottom-dwTop);

					if(dwBackTop >= dwT)
						pWall->GetMaterial()->DrawSliver(pDest, dwColumn, dwT, dwBackTop, fWallPos/pWall->m_fLength, fV1, 1.0f);

					if(dwBackBottom <= dwB)
						pWall->GetMaterial()->DrawSliver(pDest, dwColumn, dwBackBottom, dwB, fWallPos/pWall->m_fLength, 0.0f, fV2);
				}

				pBackSector->Draw(pDest, dwColumn,
					MAX(MAX(dwTop, dwBackTop), dwCH),
					MIN(MIN(dwBottom, dwBackBottom), dwFH),
					fvRay, dwBounceCount);
			}
		}
		else
		{
			gbInt dwTop = pDest->h/2-(int)((pDest->h*(m_fCeilingHeight+m_fCeilingSlope*fSlopePos))/fMinDist);
			gbInt dwBottom = pDest->h/2-(int)(pDest->h*(m_fFloorHeight+m_fFloorSlope*fSlopePos)/fMinDist);
			dline(pDest, dwColumn, dwCH, dwTop, m_pCeilingMaterial?m_pCeilingMaterial->GetColor():0);
			if(pWall->GetMaterial())
			{
				if(pWall == m_pvWalls[1]  &&  dwBounceCount < 4)
				{
					gbInt dwT = dwTop;
					gbInt dwB = dwBottom;
					dwT = MAX(dwT, dwCH);
					dwB = MIN(dwB, dwFH);

					gbFloat fWX = pWall->m_pStart->GetX();
					gbFloat fWY = pWall->m_pStart->GetY();
					gbFloat fDot = ((fvRay[0]-fWX)*pWall->m_fvNormal[0]+(fvRay[1]-fWY)*pWall->m_fvNormal[1])*2.0f;
					gbFloat fDot2 = (fvRay[2]*pWall->m_fvNormal[0]+fvRay[3]*pWall->m_fvNormal[1])*2.0f;
					gbFloat fvMirrorRay[4] = {
						fvRay[0]-pWall->m_fvNormal[0]*fDot,
						fvRay[1]-pWall->m_fvNormal[1]*fDot,
						fvRay[2]-pWall->m_fvNormal[0]*fDot2,
						fvRay[3]-pWall->m_fvNormal[1]*fDot2
					};

					Draw(pDest, dwColumn, dwT, dwB, fvMirrorRay, dwBounceCount+1);
					drawing_mode(DRAW_MODE_TRANS, NULL, 0, 0);
					vline(pDest, dwColumn, dwT, dwB, 0);
					solid_mode();
				}
				else
				{
					gbInt dwT = dwTop;
					gbInt dwB = dwBottom;
					dwT = MAX(dwT, dwCH);
					dwB = MIN(dwB, dwFH);
	
					gbFloat fV1 = (dwT-dwTop)/(gbFloat)(dwBottom-dwTop);
					gbFloat fV2 = (dwB-dwTop)/(gbFloat)(dwBottom-dwTop);
					pWall->GetMaterial()->DrawSliver(pDest, dwColumn, dwT, dwB, fWallPos/pWall->m_fLength, fV1, fV2);
				}
			}
			uline(pDest, dwColumn, dwFH, dwBottom, m_pFloorMaterial?m_pFloorMaterial->GetColor():0);
		}
	}
}
