#include <cmath>
#define ALLEGRO_FIX_H  // Don't use Allegro's fix class
#include <allegro.h>
#include <winalleg.h>
#include "Duke.h"

dkWall::dkWall() :
	m_pStart(NULL),
	m_pEnd(NULL),
	m_pBackWall(NULL),
	m_pMaterial(NULL),
	m_pPrev(this),
	m_pNext(this)
{
}

dkWall::~dkWall()
{
}

void dkWall::SetVertices(dkVertex_p pStart, dkVertex_p pEnd)
{
	m_pStart = pStart;
	m_pEnd = pEnd;

	m_fvDirection[0] = m_pEnd->GetX()-m_pStart->GetX();
	m_fvDirection[1] = m_pEnd->GetY()-m_pStart->GetY();
	m_fLength = sqrt(m_fvDirection[0]*m_fvDirection[0]+m_fvDirection[1]*m_fvDirection[1]);
	m_fvDirection[0] /= m_fLength;
	m_fvDirection[1] /= m_fLength;

	m_fvNormal[0] = -m_fvDirection[1];
	m_fvNormal[1] = m_fvDirection[0];
}

void dkWall::Join(dkWall* pBackWall)
{
	m_pBackWall = pBackWall;
	pBackWall->m_pBackWall = this;
}

gbFloat dkWall::PointPosition(gbFloat fX, gbFloat fY)
{
	return (fX-m_pStart->GetX())*m_fvDirection[0]+(fY-m_pStart->GetY())*m_fvDirection[1];
}

gbFloat dkWall::PointOffset(gbFloat fX, gbFloat fY)
{
	return (fX-m_pStart->GetX())*m_fvNormal[0]+(fY-m_pStart->GetY())*m_fvNormal[1];
}

void dkWall::PointClosest(gbFloat fX, gbFloat fY, gbFloat& fOutX, gbFloat& fOutY)
{
	gbFloat fPos = PointPosition(fX, fY);
	fOutX = m_pStart->GetX()+m_fvDirection[0]*fPos;
	fOutY = m_pStart->GetY()+m_fvDirection[1]*fPos;
}

gbFloat dkWall::RayIntersect(gbFloat fX, gbFloat fY, gbFloat fDX, gbFloat fDY, gbFloat& fOutX, gbFloat& fOutY)
{
	gbFloat fDist = -PointOffset(fX, fY)/(m_fvNormal[0]*fDX+m_fvNormal[1]*fDY);
	fOutX = fX+fDX*fDist;
	fOutY = fY+fDY*fDist;

	return fDist;
}
